#include <GestaltEqu.h>
#include "Useful.h"
#include "Essential.h"

//	The Essence-Shell Hack

//	Allan and Scott wrote this one day in June, not long before MacHack 92

//		jbx		6/28/92		Translated for use in Think C.  This file contains
//							the patches themselves, relying on a different
//							code resource to actually load them.

//		arf		6/16/92		Fixed up the menuHook stuff.  Patching InitMenus was bogus
//							because we were only storing one old MenuHook value, but each
//							app gets its own copy, and wed have to maintain one per
//							process.  That would be lame.  Now well hook in right
//							as the user clicks in the menu bar, then unhook when theyre done.
//		stb		6/16/92		Showed this to a bunch of guys who had next to nothing
//							to say.  Gee, thanks, guys :-)  Normalized spelling,
//							and added more comments.
//		stb		6/16/92		Whoops, InitMenus clears MenuHook.  Time to tail-patch InitMenus!
//		stb		6/16/92		Hook into MenuHook.
//		stb		6/16/92		Merge in Allans Gestalt code.  Its hard to get used
//							to this three hour time difference!
//		stb		6/15/92		Add alternate icon drawing; try drawing all the way
//							across the menu bar instead of a simple rect at the left.
//							Hmm, it fits better, but its harder to see it all at once.
//		stb		6/14/92		Change some names for clarity

void DrawTheCounters(void);

pascal short EssentialGestalt(OSType selector,long *response);

typedef pascal long MenuSelectRoutine(Point pt);
pascal long pMenuSelect(Point pt, MenuSelectRoutine *oldMS);

typedef pascal Handle GetResourceRoutine(ResType rType, short ID);
pascal Handle pGetResource(ResType rType, short ID, GetResourceRoutine *oldGR);

void pLoadSeg(void);

void pNewHandle(void);
void pDisposHandle(void);
void pNewPtr(void);
void pDisposPtr(void);

void pRead(void);
void pWrite(void);

void pGNEFilter(EventRecord *er);

//
// By convention from the essential hack loader, the init resource begins with the
// four-byte gestalt selector for this hack, and an offset to the gestalt routine
// to be installed on behalf of this hack.  After that is a series of two-word
// entries, where the first word specifies either a trap to be patched or a
// low-memory vector to be patched.  When the first word becomes zero, the traps
// have ended.

void main(void) {
	asm {
		dc.l	sEssential	; our gestalt selector.
		dc.w	EssentialGestalt

		dc.w	0xA93D
		dc.w	@_MenuSelect
		dc.w	0xA9A0
		dc.w	@_GetResource
		dc.w	0xA9F0
		dc.w	@_LoadSeg
		dc.w	0xA022
		dc.w	@_NewHandle
		dc.w	0xA023
		dc.w	@_DisposHandle
		dc.w	0xA01E
		dc.w	@_NewPtr
		dc.w	0xA01F
		dc.w	@_DisposPtr
		dc.w	0xA002
		dc.w	@_Read
		dc.w	0xA003
		dc.w	@_Write

		dc.w	666		; this is the address of JGNEFilter.  make of it what you will...
		dc.w	@_GNEFilter
		dc.w	0

	// OK, enough with the header.  here's the patches.  First, MenuSelect.  We
	// want to tail-patch MenuSelect, so our patch inserts an extra parameter
	// (the old trap address) so our patch routine can do everything at a high
	// level.  Note that this requires the trap routine (pMenuSelect) to be
	// declared as a pascal routine, so that it automatically clears away
	// the extra parameter we added.

_MenuSelect:
		dc.l	WayBadValue
		move.l	(a7),-(a7)			; make space for an extra parameter.
		move.l	@_MenuSelect,4(a7)	; the old address is that parameter.
		bra		pMenuSelect			; go to high-level stuff.

	// We do the same thing for GetResource, which we also want to tail-patch.
_GetResource:
		dc.l	WayBadValue
		move.l	(a7),-(a7)			; make space for an extra parameter.
		move.l	@_GetResource,4(a7)	; the old address is that parameter.
		bra		pGetResource		; go to high-level stuff.

	// We just want to head-patch LoadSeg, so we set up a4, call the high-level
	// patch, and when it returns we go to the real routine.  Note that all the
	// registers are saved here even though many of the patches don't modify any
	// of them.  The idea is, might as well be safe, and leave the head patches to
	// do everything at a high level without having to worry about register saving,
	// which the tail patch stuff has to do.
_LoadSeg:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pLoadSeg
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_LoadSeg,-(a7)
		rts

	// same for NewHandle.
_NewHandle:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pNewHandle
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_NewHandle,-(a7)
		rts

	// same for DisposHandle.
_DisposHandle:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pDisposHandle
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_DisposHandle,-(a7)
		rts

	// and same for NewPtr
_NewPtr:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pNewPtr
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_NewPtr,-(a7)
		rts

	// and DisposPtr
_DisposPtr:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pDisposPtr
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_DisposPtr,-(a7)
		rts

	// and Read
_Read:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pRead
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_Read,-(a7)
		rts

	// and Write
_Write:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		bsr		pWrite
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_Write,-(a7)
		rts

_GNEFilter:
		dc.l	WayBadValue
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		move.l	a1,-(a7)	; a1 points to the event being filtered.
		bsr		pGNEFilter
		addq	#4,a7
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
		move.l	@_GNEFilter,-(a7)
		bne.s	@norm
		addq	#4,a7		; don't go to the old filter if it was zero!
norm:;	rts
	}
}

	// oldMenuHook contains the old value of MenuHook except that if the old value
	// was 0 or -1, meaning no hook, this value points at pNullHook.  This
	// way we can always jump through oldMenuHook and not crash.
static ProcPtr oldMenuHook;

	// a routine that does nothing can be useful.
static void pNullHook(void) {
}

	// pMenuHook does our stuff and then falls through to the old hook.
static void pMenuHook(void) {
	asm {
		subq	#4,a7	; reserve space for old menu hook value
		movem.l	d0/d1/d2/a0/a1/a4,-(a7)
		lea		main,a4
		move.l	oldMenuHook,24(a7)	; stuff the space so we return to the original hook
	}

	DrawTheCounters();

	asm {
		movem.l	(a7)+,d0/d1/d2/a0/a1/a4
	}
	// we return to the old menuhook here.
}

// this routine does everything for MenuSelect, both our stuff and what the old
// trap did.  we start out by saving the registers because we use some of them and
// we want to preserve as many of them as possible.  In reality the call to oldMS
// trashes a0 which happens to be OK for this trap because it trashes a0 anyway, but
// this is not ideal.
pascal long pMenuSelect(Point pt, MenuSelectRoutine *oldMS) {
	long             regs[6];
	long             result;
	register ProcPtr realOldMenuHook;

	asm { movem.l d0/d1/d2/a0/a1/a4,regs }	// save the registers.
	SetA4(&main); // set up access to our globals.

	oldMenuHook = realOldMenuHook = MenuHook;
	MenuHook = &pMenuHook;
	if (realOldMenuHook == 0 || realOldMenuHook == (ProcPtr) -1) {
		oldMenuHook = &pNullHook;
	}

	asm { movem.l regs,d0/d1/d2/a0/a1/a4 } // restore the registers.

	result = oldMS(pt);	// call the old MenuSelect.  This line trashes a0!

	MenuHook = realOldMenuHook;	// this doesn't disturb the registers and doesn't use globals.

	return result;
}

// this structure contains all our counters.  We declare them as a structure because
// this is the only way we can be assured about their layout in memory.

struct {
	long	LoadSeg,            LoadSegRT;
	long	LoadSegHitDisk,     LoadSegHitDiskRT;
	long	GetResource,        GetResourceRT;
	long	GetResourceHitDisk, GetResourceHitDiskRT;
	long	NewHandle,          NewHandleRT;
	long	DisposHandle,       DisposHandleRT;
	long	NewPtr,             NewPtrRT;
	long	DisposPtr,          DisposPtrRT;
	long	Read,               ReadRT;
	long	Write,              WriteRT;
} Counters;

// this bitmap points at the counters for the purposes of a copybits.

static BitMap	counterMap = {
	(void *)WayBadValue,	// the base address has to be filled in at runtime.
	512,	// this rowBytes value is way too high but that's OK we only have 1 line.
	0, 0, 1, sizeof(Counters)*8	// this is exactly the right size.
};

// the total lack of register-saving code is possible because for cases of head patching
// all the niggly register details can be done at asm level above.
short inLoadSeg;
void pLoadSeg(void) {
	Counters.LoadSeg++;
	inLoadSeg = TRUE;	// we set this back to false in GetResource.
}

// again this tail-patch routine performs the entire burden of GetResource, and
// again trashing a0 in the call to oldGR is suboptimal.  And again it's OK
// because the trap trashes a0 anyway.

short inGetResource;
pascal Handle pGetResource(ResType rType, short ID, GetResourceRoutine *oldGR) {
	long	regs[6];
	Handle	result;

	asm { movem.l d0/d1/d2/a0/a1/a4,regs }	// save the registers.
	SetA4(&main);

	Counters.GetResource++;
	inGetResource = TRUE;

	asm { movem.l regs,d0/d1/d2/a0/a1/a4 }	// restore the registers.

	result = oldGR(rType, ID);	// call the old GetResource.

	asm { movem.l d0/d1/d2/a0/a1/a4,regs }	// save the registers.
	SetA4(&main);

	inGetResource = FALSE;
	inLoadSeg = FALSE;

	asm { movem.l regs,d0/d1/d2/a0/a1/a4 }	// restore the registers.

	return result;
}

void pNewHandle(void) {
	Counters.NewHandle++;
}

void pDisposHandle(void) {
	Counters.DisposHandle++;
}

void pNewPtr(void) {
	Counters.NewPtr++;
}

void pDisposPtr(void) {
	Counters.DisposPtr++;
}

void pRead(void) {
	// Count Reads, but also decide whether GetResource or LoadSeg
	// hit the disk, and count for them.

	Counters.Read++;
	if (inLoadSeg) {
		Counters.LoadSegHitDisk++;
		inLoadSeg = 0;			// don't count more than once!
	}
	if (inGetResource) {
		Counters.GetResourceHitDisk++;
		inGetResource = 0;		// don't count more than once!
	}
}

void pWrite(void) {
	Counters.Write++;
}

//
//		This Function clears the counters, and updates the running totals...
//

static void ClearTheCounters(void) {
	Counters.LoadSegRT           += Counters.LoadSeg;           Counters.LoadSeg = 0;
	Counters.LoadSegHitDiskRT    += Counters.LoadSegHitDisk;    Counters.LoadSegHitDisk = 0;
	Counters.GetResourceRT       += Counters.GetResource;       Counters.GetResource = 0;
	Counters.GetResourceHitDiskRT+= Counters.GetResourceHitDisk;Counters.GetResourceHitDisk = 0;
	Counters.NewHandleRT         += Counters.NewHandle;         Counters.NewHandle = 0;
	Counters.DisposHandleRT      += Counters.DisposHandle;      Counters.DisposHandle = 0;
	Counters.NewPtrRT            += Counters.NewPtr;            Counters.NewPtr = 0;
	Counters.DisposPtrRT         += Counters.DisposPtr;         Counters.DisposPtr = 0;
	Counters.ReadRT              += Counters.Read;              Counters.Read = 0;
	Counters.WriteRT             += Counters.Write;             Counters.Write = 0;
}

long GNEFilterTimer, GNEFilterTimerRT;
void pGNEFilter(EventRecord *er) {
		// the first thing we do is test to see that we are taking up time from
		// a null event.  The idea is, if the user's clicking and typing, we
		// don't want to slow anything down.  Side note: if you're interested
		// in writing a patch which intercepts key-downs or mouse-downs, the
		// GNEFilter is NOT the place to do it!  The lesser reason is that GNE
		// has a weird register-based interface that a lot of other INITs mis-use
		// so you can't always tell whether to pay attention to an event or not
		// (you end up just paying attention to the what field).  The greater reason
		// is that JGNE doesn't filter all events; in particular events going to desk
		// accessories get screened out by SystemEvent first!  So, if you're looking
		// to get all keyDowns as they come through, patch SystemEvent and set the
		// what field to nullEvent for events you want to handle yourself.  In case
		// you were thinking of patching PostEvent, SystemEvent is infinitely better
		// because SystemEvent can move memory whereas PostEvent is called at interrupt
		// time.
	if (er->what == nullEvent) {
		if (er->when - GNEFilterTimer >= redrawInterval) {
			GNEFilterTimer = er->when;
			DrawTheCounters();
			ClearTheCounters();
		}
	}
}

void DrawTheCounters(void) {
	GrafPtr			oldPort;
	RgnHandle		rgn;
	GrafPtr			WMP = WMgrPort;


		// save the old port 'cause we're gonna change it.
	GetPort(&oldPort);
		// draw into the window manager port.
	SetPort(WMP);

		// make a new region for our use
	rgn = NewRgn();
		// put the current clip in it (we have to change the clip)
	GetClip(rgn);
		// make the clip region wide-open.
	ClipRect(&WMP->portBits.bounds);

	counterMap.baseAddr = (void *)&Counters;
	CopyBits(&counterMap, &WMP->portBits,
	         &counterMap.bounds, &counterMap.bounds, srcCopy, 0);

		// set the clip back.
	SetClip(rgn);
		// and get rid of our region now that we don't need it anymore.
	DisposeRgn(rgn);

	SetPort(oldPort);
}

// 	This is the routine called by Gestalt to return the Proc Pointer.
//
//	OSErr gestaltSelectorProc(OSType selector,long *response);
//
//	Note that this is not the same tack taken by Scott & Alan!  The idea
//	here is that rather than returning the address of a different routine
//	which returns other results, we have only this one routine and it
//	returns a multitude of different things, depending on the selector.
//	So, you call Gestalt with the Sntl selector to get the address of this
//	routine, and then you call this routine with other selectors to get
//	(or do) other things.

pascal short EssentialGestalt(OSType selector, long *response) {
	void	*result;
	short	err = noErr;

	void *oldA4 = SetA4(&main);
	switch(selector) {
		case sEssential:
			result = &EssentialGestalt;
		break;

		case sClearTable:
				// this selector clears the counters and then returns their address.
			ClearTheCounters();
				// fall-through to sTablePointer intentional!!!

		case sTablePointer:
				// this selector just returns the address of our counters.
			result = &Counters;
		break;

		case sUninstall:
				// this selector de-installs all our patches.
				// as you can see, this part isn't finished yet.
			SysBeep(5);
			result = 0;
		break;

		default:
			err = gestaltUnknownErr;
		break;
	};
	SetA4(oldA4);

	if (err == noErr) {
		*response = (long) result;
	}
	return err;
}

// For example, if you wanted to make a small routine which reset
// the counters, you'd do it like this:

#if 0
void ClearTheCounters(void) {
	pascal short (*Essential)(OSType selector, long *response);
	long	*tables;

	// first we test for the existence of the INIT and get the address of
	// EssentialGestalt all at once:
	if (Gestalt(sEssential, &Essential) == noErr) {
		// Then we call pGestalt as we wish.
		if (Essential(sClearTable, (void *)&tables) == noErr) {
			// ordinarily you wouldn't vandalize the tables like this but we want to
			// know it worked.
			*tables++ = 0xAAAAAAAA;
		}
	}
}
#endif